<?php

namespace JSONRPC;

/**
 * JSONRPCException The JSON RPC Exception class
 */
class JSONRPCException extends \Exception
{
}

/**
 * Client The JSON RPC client class
 *
 *
 *
 * @version 1.1
 * @author sathya
 */
class Client
{
    /**
     * IP address of the JSON RPC server
     * @var string
     */
    private $host = "";

    /**
     * Port number of the JSON RPC server
     * @var int
     */
    private $port = 2001;

    /**
     * Enable SSL
     * @var bool
     */
    private $ssl = false;

    /**
     * Enable certificate verification
     * @var bool
     */
    private $sslVerifyPeer = true;

    /**
     * Enable certificate verification
     * @param $value bool
     */
    public function setSSLVerifyPeer($value)
    {
        $this->sslVerifyPeer = $value;
    }

    /**
     * Path to the certificate cuthority's certificate
     * @var string
     */
    private $caFile = "";

    /**
     * Set the path to the certificate cuthority's certificate
     * @param $value string
     */
    public function setCAFile($value)
    {
        $this->caFile = $value;
    }

    /**
     * Username for basic auth
     * @var string
     */
    private $username = "";

    /**
     * Set username for basic auth
     * @param $value string
     */
    public function setUsername($value)
    {
        $this->username = $value;
    }

    /**
     * Password for basic auth
     * @var string
     */
    private $password = "";

    /**
     * Set password for basic auth
     * @param $value string
     */
    public function setPassword($value)
    {
        $this->password = $value;
    }

    /**
     * Holds the socket connection
     * @var socket
     */
    private $socket = null;

    /**
     * Default constructor
     * @param $host string IP address of the JSON RPC server
     * @param $port int Port number of the JSON RPC server
     * @param $ssl bool Enable SSL
     */
    public function __construct($host, $port = 2001, $ssl = false)
    {
        $this->host = $host;
        $this->port = $port;
        $this->ssl = $ssl;

        $this->connect();
    }

    public function __destruct()
    {
        if($this->socket) @fclose($this->socket);
        $this->socket = null;
    }

    /**
     * Connects to Homegear
     */
    private function connect()
    {
        if(!$this->socket)
        {
            $this->socket = @stream_socket_client("tcp://".$this->host.":".$this->port, $errorNumber, $errorString, 10);
            if(!$this->socket) throw new JSONRPCException("Could not open socket. Host: ".$this->host." Port: ".$this->port." Error: $errorString ($errorNumber)");
            if($this->ssl)
            {
                stream_set_blocking($this->socket, true);
                stream_context_set_option($this->socket, 'ssl', 'SNI_enabled', true);
                if($this->caFile) stream_context_set_option($this->socket, 'ssl', 'cafile', $this->caFile);
                stream_context_set_option($this->socket, 'ssl', 'verify_peer', $this->sslVerifyPeer);
                $secure = stream_socket_enable_crypto($this->socket, true, STREAM_CRYPTO_METHOD_TLS_CLIENT);
                if(!$secure)
                {
                    @fclose($this->socket);
                    throw new JSONRPCException("JSONRPC error: Failed to enable SSL.");
                }
                stream_set_blocking($this->socket, false);
            }
        }
    }

    /**
     * Sends a JSON RPC request to JSON RPC server
     * @param $request string Valid JSON-RPC request
     * @return string The returned JSON string
     */
    private function sendRequest($request)
    {
        $response = '';
        $retries = 0;
        $startTime = time();
        while(!$response && $retries < 20)
        {
            if(!$this->socket) $this->connect();

            $response = '';
            $query = "POST / HTTP/1.1\r\nUser_Agent: HG-JSONRPC-Client\r\nHost: ".$this->host."\r\nConnection: Keep-Alive\r\nContent-Type: application/json\r\n";

            if($this->username)
            {
                $query .= "Authorization: Basic ".base64_encode($this->username.":".$this->password)."\r\n";
            }

            $query .= "Content-Length: ".strlen($request)."\r\n\r\n".$request."\r\n";

            $bytesWritten = 0;
            $continueLoop = false;
            $querySize = strlen($query);
            while($bytesWritten < $querySize)
            {
                $result = @fputs($this->socket, $query, 1024);
                if (!$result)
                {
                    if($retries == 19) throw new JSONRPCException("Error sending data to server.");
                    else
                    {
                        @fclose($this->socket);
                        $this->socket = null;
                        $retries++;
                        usleep(50);
                        $continueLoop = true;
                        break;
                    }
                }
                $bytesWritten += $result;
                $query = substr($query, $result);
            }
            if($continueLoop) continue;

            while (!feof($this->socket) && (time() - $startTime) < 30)
            {
                $response .= @fgets($this->socket);
                $pos = strpos($response, "\r\n\r\n");
                $contentLength = 0;
                if ($pos !== false) {
                    //Header complete
                    if(strncmp($response, "HTTP/1.1 200 OK", 15) !== 0) throw new JSONRPCException("JSON-RPC error:\r\n".$response);
                    $contentLengthPos = strpos($response, 'Content-Length: ');
                    if ($contentLengthPos === false) throw new JSONRPCException("Invalid response received.");
                    $contentLengthPos2 = strpos($response, "\r\n", $contentLengthPos);
                    $contentLength = intval(substr($response, $contentLengthPos + 16, $contentLengthPos2 - ($contentLengthPos + 16)));
                    if ($pos + 4 + $contentLength === strlen($response)) {
                        //Packet complete
                        $response = substr($response, $pos + 4);
                        break;
                    }
                }
            }

            if(!$response)
            {
                @fclose($this->socket);
                $this->socket = null;
            }
            $retries++;
        }
        if(!$response) throw new JSONRPCException("JSONRPC error: Response was empty.");
        return $response;
    }

    /**
     * Sends an JSON RPC request to the JSON RPC server
     * @param $methodName string Name of the JSON RPC method to call
     * @param $params array Array of parameters to pass to the JSON RPC method. Type is detected automatically. For a struct just encode an array within $params.
     * @return array Array of returned parameters
     */
    public function send($methodName, $params)
    {
        $data = [
            'jsonrpc' => '2.0',
            'id' => 1,
            'method' => $methodName,
            'params' => $params
        ];
        $request = json_encode($data);
        $response = $this->sendRequest($request);
        $response = json_decode($response, true);
        if (array_key_exists('error', $response)) throw new JSONRPCException($response['error']['message']);
        return $response['result'];
    }

    public function __call($methodName, $params)
    {
        return $this->send($methodName, $params);
    }
}
